package ssw.mj.test;

import ssw.mj.symtab.Obj;
import ssw.mj.symtab.Struct;
import ssw.mj.symtab.Tab;

/**
 * Test cases for the symbol table.
 * 
 * @author AW
 */
public class SymTabTest extends IdentifiedTestCase {
	
	protected void setUp () { 
		super.setUp();
		
		Tab.init(); 
	}
	
	public void testUniverse () {
		Obj actual, expected, locals;
		
		actual = Tab.find("int");
		expected = new Obj(Obj.Type, "int", Tab.intType);
		assertEquals("Universe 'int'", expected, actual);
		
		actual = Tab.find("char");
		expected = new Obj(Obj.Type, "char", Tab.charType);
		assertEquals("Universe 'char'", expected, actual);
		
		actual = Tab.find("null");
		expected = new Obj(Obj.Con, "null", Tab.nullType);
		expected.val = 0;
		assertEquals("Universe 'null'", expected, actual);
		
		actual = Tab.find("chr");
		expected = new Obj(Obj.Var, "i", Tab.intType);
		expected.adr = 0;
		expected.level = 1;
		assertEquals("Param 'i'", expected, actual.locals);
		locals = expected;
		expected = new Obj(Obj.Meth, "chr", Tab.charType);
		expected.adr = 0;
		expected.nPars = 1;
		expected.locals = locals;
		assertEquals("Universe 'chr'", expected, actual);
		assertEquals("Universe 'chrObj'", expected, Tab.chrObj);
		assertSame("Universe 'chrObj' found", Tab.chrObj, actual);

		actual = Tab.find("ord");
		locals = new Obj(Obj.Var, "ch", Tab.charType);
		locals.adr = 0;
		locals.level = 1;
		expected = new Obj(Obj.Meth, "ord", Tab.intType);
		expected.adr = 0;
		expected.nPars = 1;
		expected.locals = locals;
		assertEquals("Universe 'ord'", expected, actual);
		assertEquals("Universe 'ordObj'", expected, Tab.ordObj);
		assertSame("Universe 'ordObj' found", Tab.ordObj, actual);
		
		actual = Tab.find("len");
		locals = new Obj(Obj.Var, "arr", new Struct(Struct.Arr, Tab.noType));
		locals.adr = 0;
		locals.level = 1;
		expected = new Obj(Obj.Meth, "len", Tab.intType);
		expected.adr = 0;
		expected.nPars = 1;
		expected.locals = locals;
		assertEquals("Universe 'len'", expected, actual);		
		assertEquals("Universe 'lenObj'", expected, Tab.lenObj);
		assertSame("Universe 'lenObj' found", Tab.lenObj, actual);
	}
	
	public void testInsertIntoUniverse () {
		Obj actual, expected;
		
		actual = Tab.insert(Obj.Prog, "MyClass", Tab.noType);
		expected = new Obj(Obj.Prog, "MyClass", Tab.noType);
		assertEquals("Insert 'MyClass'", expected, actual);

		actual = Tab.find("MyClass");
		assertEquals("Find 'MyClass'", expected, actual);
	}
	
	public void testInsertGlobal () {
		Obj actual, expected;

		Tab.openScope();	// scope for global variables
		
		actual = Tab.insert(Obj.Var, "x0", Tab.intType);
		expected = new Obj(Obj.Var, "x0", Tab.intType);
		expected.adr = 0;
		expected.level = 0;
		assertEquals("First global", expected, actual);

		actual = Tab.find("x0");
		assertEquals("Find first global", expected, actual);

		actual = Tab.insert(Obj.Var, "x1", Tab.intType);
		expected = new Obj(Obj.Var, "x1", Tab.intType);
		expected.adr = 1;
		expected.level = 0;
		assertEquals("Second global", expected, actual);
	}

	///////////////////////////////////////////////////////////////
	// WARNING:                                                  //
	// on my machine (1GHz/256MB) this takes about 3 minutes !!! //
	// If you want to run it, simply comment out the fail(...);  //
	// Run it at least once (in a while).                        //
	///////////////////////////////////////////////////////////////
	public void testInsertMaxGlobals () {
		Obj actual, expected;

		fail(">>> To speed up testing testInsertMaxGlobals was not run! <<<");
		
		Tab.openScope();  // scope for global variables

		actual = Tab.insert(Obj.Var, "x0", Tab.intType);
		expected = new Obj(Obj.Var, "x0", Tab.intType);
		expected.adr = 0;
		expected.level = 0;
		assertEquals("First global", expected, actual);

		actual = Tab.find("x0");
		assertEquals("Find first global", expected, actual);

		for (int i = 1; i < 32768; i++) 
			actual = Tab.insert(Obj.Var, "x" + i, Tab.intType);

		expected = new Obj(Obj.Var, "x32767", Tab.intType);
		expected.adr = 32767;
		expected.level = 0;
		assertEquals("Last valid global", expected, actual);

		actual = Tab.find("x32767");
		assertEquals("Find last valid global", expected, actual);

		actual = Tab.insert(Obj.Var, "x32768", Tab.intType);
		expected = Tab.noObj;
		assertEquals("Insert one more than allowed", expected, actual);
	}
	
	public void testInsertLocal () {
		Obj actual, expected;

		Tab.openScope();	// global scope
		Tab.openScope();	// local scope, level 1
		
		actual = Tab.insert(Obj.Var, "x", Tab.intType);
			
		expected = new Obj(Obj.Var, "x", Tab.intType);
		expected.adr = 0;
		expected.level = 1;
		assertEquals("First valid local var on level 1", expected, actual);

		actual = Tab.find("x");
		assertEquals("Find first valid local var on level 1", expected, actual);
		
		actual = Tab.insert(Obj.Var, "y", Tab.intType);
		expected = new Obj(Obj.Var, "y", Tab.intType);
		expected.adr = 1;
		expected.level = 1;
		assertEquals("Second valid local var on level 1", expected, actual);		
	}

	public void testInsertMaxLocals () {
		Obj actual, expected;

		Tab.openScope();	// global scope
		Tab.openScope();	// local scope, level 1
		
		for (int i = 0; i < 127; i++) 
			Tab.insert(Obj.Var, "x" + i, Tab.intType);
		actual = Tab.insert(Obj.Var, "x127", Tab.intType);
					
		expected = new Obj(Obj.Var, "x127", Tab.intType);
		expected.adr = 127;
		expected.level = 1;
		assertEquals("Last valid local var on level 1", expected, actual);
		
		actual = Tab.find("x127");
		assertEquals("Find last valid local var on level 1", expected, actual);
		
		actual = Tab.insert(Obj.Var, "x128", Tab.intType);
		expected = Tab.noObj;
		assertEquals("Insert one more local var on level 1 than allowed", expected, actual);
	}
	
	public void testAssignableTo () {
		assertTrue("int = int", Tab.intType.assignableTo(Tab.intType));
		assertTrue("char = char", Tab.charType.assignableTo(Tab.charType));
		assertFalse("char = int", Tab.intType.assignableTo(Tab.charType));
		assertFalse("int = char", Tab.charType.assignableTo(Tab.intType));
		
		Struct t1, t2, lenPar = new Struct(Struct.Arr, Tab.noType);

		t1 = new Struct(Struct.Arr, Tab.intType);
		t2 = new Struct(Struct.Arr, Tab.intType);
		assertTrue("[int] = [int] (1)", t1.assignableTo(t2));
		assertTrue("[int] = [int] (2)", t2.assignableTo(t1));
		assertTrue("[int] = null", Tab.nullType.assignableTo(t1));
		assertTrue("len([int])", t1.assignableTo(lenPar));
		t1 = new Struct(Struct.Arr, Tab.charType);
		t2 = new Struct(Struct.Arr, Tab.charType);
		assertTrue("[char] = [char] (1)", t1.assignableTo(t2));
		assertTrue("[char] = [char] (2)", t2.assignableTo(t1));
		assertTrue("[char] = null", Tab.nullType.assignableTo(t1));	
		assertTrue("len([char])",  t1.assignableTo(lenPar));
	}
	
	public static void main (String[] args) {
		junit.textui.TestRunner.run(SymTabTest.class);
	}
}
